import { parser } from 'lezer-python';
import { LezerLanguage, indentNodeProp, continuedIndent, foldNodeProp } from '@codemirror/next/language';
import { styleTags, tags } from '@codemirror/next/highlight';

/// A language provider based on the [Lezer Python
/// parser](https://github.com/lezer-parser/python), extended with
/// highlighting and indentation information.
const pythonLanguage = LezerLanguage.define({
    parser: parser.configure({
        props: [
            indentNodeProp.add({
                Body: continuedIndent()
            }),
            foldNodeProp.add({
                Body(tree) { return { from: tree.from + 1, to: tree.to - 1 }; },
                ArrayExpression(tree) { return { from: tree.from + 1, to: tree.to - 1 }; },
                DictionaryExpression(tree) { return { from: tree.from + 1, to: tree.to - 1 }; }
            }),
            styleTags({
                "async '*' '**' FormatConversion": tags.modifier,
                "for while if elif else try except finally return raise break continue with pass assert await yield": tags.controlKeyword,
                "in not and or is del": tags.operatorKeyword,
                "import from def class global nonlocal lambda": tags.definitionKeyword,
                "with as print": tags.keyword,
                self: tags.self,
                Boolean: tags.bool,
                None: tags.null,
                VariableName: tags.variableName,
                PropertyName: tags.propertyName,
                Comment: tags.lineComment,
                Number: tags.number,
                String: tags.string,
                FormatString: tags.special(tags.string),
                UpdateOp: tags.updateOperator,
                ArithOp: tags.arithmeticOperator,
                BitOp: tags.bitwiseOperator,
                CompareOp: tags.compareOperator,
                AssignOp: tags.definitionOperator,
                Ellipsis: tags.punctuation,
                At: tags.meta,
                "( )": tags.paren,
                "[ ]": tags.squareBracket,
                "{ }": tags.brace,
                ".": tags.derefOperator,
                ", ;": tags.separator
            })
        ],
    }),
    languageData: {
        closeBrackets: { brackets: ["(", "[", "{", "'", '"', "'''", '"""'] },
        commentTokens: { line: "#" },
        indentOnInput: /^\s*[\}\]\)]$/
    }
});
/// Returns an extension that installs the Python language.
function python() {
    return pythonLanguage;
}

export { python, pythonLanguage };
